/*
 *   LADCCA
 *    
 *   Copyright (C) 2002 Robert Ham (node@users.sourceforge.net)
 *    
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>

#include <alsa/asoundlib.h>

#include "alsa_patch.h"

void
alsa_patch_init (alsa_patch_t * patch)
{
  int err;
  
  err = snd_seq_port_subscribe_malloc (&patch->sub);
  if (err)
    abort ();
  
  uuid_clear (patch->from_id);
  uuid_clear (patch->to_id);
}

void
alsa_patch_free (alsa_patch_t * patch)
{
  snd_seq_port_subscribe_free (patch->sub);
}

alsa_patch_t *
alsa_patch_new ()
{
  alsa_patch_t * patch;
  patch = cca_malloc (sizeof (alsa_patch_t));
  alsa_patch_init (patch);
  return patch;
}

alsa_patch_t *
alsa_patch_new_with_sub (const snd_seq_port_subscribe_t * sub)
{
  alsa_patch_t * patch;
  patch = alsa_patch_new ();
  alsa_patch_set_sub (patch, sub);
  return patch;
}

alsa_patch_t *
alsa_patch_new_with_query (const snd_seq_query_subscribe_t * query)
{
  alsa_patch_t * patch;
  patch = alsa_patch_new ();
  alsa_patch_set_sub_from_query (patch, query);
  return patch;
}

void
alsa_patch_destroy (alsa_patch_t * patch)
{
  alsa_patch_free (patch);
  free (patch);
}

void
alsa_patch_get_from_id (alsa_patch_t * patch, uuid_t id)
{
  uuid_copy (id, patch->from_id);
}

void
alsa_patch_get_to_id   (alsa_patch_t * patch, uuid_t id)
{
  uuid_copy (id, patch->to_id);
}

snd_seq_port_subscribe_t *
alsa_patch_get_sub (alsa_patch_t * patch)
{
  return patch->sub;
}

int
alsa_patch_from_id_is_null (alsa_patch_t * patch)
{
  return uuid_is_null (patch->from_id);
}

int
alsa_patch_to_id_is_null   (alsa_patch_t * patch)
{
  return uuid_is_null (patch->to_id);
}

void
alsa_patch_set_from_id (alsa_patch_t * patch, uuid_t id)
{
  uuid_copy (patch->from_id, id);
}

void
alsa_patch_set_to_id   (alsa_patch_t * patch, uuid_t id)
{
  uuid_copy (patch->to_id, id);
}

void
alsa_patch_set_sub     (alsa_patch_t * patch, const snd_seq_port_subscribe_t * sub)
{
  if (sub)
    snd_seq_port_subscribe_copy (patch->sub, sub);
}

void
alsa_patch_set_sub_from_query (alsa_patch_t * patch, const snd_seq_query_subscribe_t * query)
{
  if (query)
    alsa_patch_set_sub (patch, query_to_port (query));
}

void
alsa_patch_set_id (uuid_t patch_id, unsigned char alsa_client_id, const project_t * project)
{
  cca_list_t * list;
  client_t * client;
  int found = 0;
  
  list = project->clients;
  while (list)
    {
      client = (client_t *) list->data;
      
      if (client_get_alsa_client_id (client) == alsa_client_id)
        {
          found = 1;
          break;
        }
      
      list = list->next;
    }
  
  if (found)
    {
      uuid_copy (patch_id, client->cca_client->id); 
    }
}

void
alsa_patch_set   (alsa_patch_t * patch, const project_t * project)
{
  const snd_seq_addr_t * addr;
  
  if (!patch->sub)
    return;
  
  addr = snd_seq_port_subscribe_get_sender (patch->sub);
  alsa_patch_set_id (patch->from_id, addr->client, project);
  
  addr = snd_seq_port_subscribe_get_dest (patch->sub);
  alsa_patch_set_id (patch->to_id, addr->client, project);
}

int
alsa_patch_find_client_id (uuid_t patch_id, const project_t * project)
{
  cca_list_t * list;
  client_t * client;
  int found = 0;
  unsigned char id;
  
  list = project->clients;
  while (list)
    {
      client = (client_t *) list->data;
      
      if (uuid_compare (patch_id, client->cca_client->id) == 0)
        {
          found = 1;
          break;
        }
      
      list = list->next;
    }
  
  if (found)
    {
      id = client_get_alsa_client_id (client);
      if (!id)
        {
          fprintf (stderr, "%s: could not unset alsa patch as client '%s' has not sent its alsa client id yet\n",
                   __FUNCTION__, cca_client_get_id_str (client->cca_client));
          return -1;
        }
      return id;
    }
  
  return 0;
}

int
alsa_patch_unset (alsa_patch_t * patch, const project_t * project)
{
  snd_seq_addr_t addr;
  const snd_seq_addr_t * addrptr;
  int client_id;
  
  if (!uuid_is_null (patch->from_id))
    {
      client_id = alsa_patch_find_client_id (patch->from_id, project);
      
      if (client_id == -1)
        return 1;
      
      if (client_id)
        {
          addrptr = snd_seq_port_subscribe_get_sender (patch->sub);
          addr.client = client_id;
          addr.port = addrptr->port;
          snd_seq_port_subscribe_set_sender (patch->sub, &addr);
          uuid_clear (patch->from_id);
        }
    }

  if (!uuid_is_null (patch->to_id))
    {
      client_id = alsa_patch_find_client_id (patch->to_id, project);

      if (client_id == -1)
        return 1;
      
      if (client_id)
        {
          addrptr = snd_seq_port_subscribe_get_dest (patch->sub);
          addr.client = client_id;
          addr.port = addrptr->port;
          snd_seq_port_subscribe_set_dest (patch->sub, &addr);
          uuid_clear (patch->to_id);
        }
    }
  
  return 0;
}

const snd_seq_port_subscribe_t *
query_to_port (const snd_seq_query_subscribe_t * query)
{
  static snd_seq_port_subscribe_t * sub = NULL;
  
  if (!query)
    return NULL;
  
  if (!sub)
    {
      int err;
      err = snd_seq_port_subscribe_malloc (&sub);
      if (err)
        abort;
    }
  
  snd_seq_port_subscribe_set_sender      (sub, snd_seq_query_subscribe_get_root (query));
  snd_seq_port_subscribe_set_dest        (sub, snd_seq_query_subscribe_get_addr (query));
  snd_seq_port_subscribe_set_exclusive   (sub, snd_seq_query_subscribe_get_exclusive (query));
  snd_seq_port_subscribe_set_time_update (sub, snd_seq_query_subscribe_get_time_update (query));
  snd_seq_port_subscribe_set_time_real   (sub, snd_seq_query_subscribe_get_time_real (query));
  snd_seq_port_subscribe_set_queue       (sub, snd_seq_query_subscribe_get_queue (query));
  
  return sub;
}

int
print_query_to_string (char * str, size_t str_size, const char * from,
                       const char * to, const snd_seq_port_subscribe_t * sub)
{
  const snd_seq_addr_t * sender;
  const snd_seq_addr_t * dest;
  
  sender = snd_seq_port_subscribe_get_sender (sub);
  dest   = snd_seq_port_subscribe_get_dest (sub);
  
  return snprintf (str, str_size, "%s %d %s %d %d %d %d %d",
                   from,
                   sender->port,
                   to,
                   sender->port,
                   snd_seq_port_subscribe_get_queue (sub),
                   snd_seq_port_subscribe_get_exclusive (sub),
                   snd_seq_port_subscribe_get_time_update (sub),
                   snd_seq_port_subscribe_get_time_real (sub));
}

const char *
alsa_patch_get_spec (alsa_patch_t * patch)
{
  static char * spec = NULL;
  static size_t spec_size = 32;
  int size;
  const snd_seq_addr_t * addr;
  char from[37];
  char to[37];
  
  if (!spec)
    spec = cca_malloc (spec_size);

  if (alsa_patch_from_id_is_null (patch))
    {
      addr = snd_seq_port_subscribe_get_sender (patch->sub);
      sprintf (from, "%d", addr->client);
    }
  else
    uuid_unparse (patch->from_id, from);

  if (alsa_patch_to_id_is_null (patch))
    {
      addr = snd_seq_port_subscribe_get_dest (patch->sub);
      sprintf (to, "%d", addr->client);
    }
  else
    uuid_unparse (patch->to_id, to);
  
  size = print_query_to_string (spec, spec_size, from, to, patch->sub);

  if (size >= spec_size)
    {
      spec_size = size + 1;
      spec = cca_realloc (spec, spec_size);
      print_query_to_string (spec, spec_size, from, to, patch->sub);
    }
  
  return spec;
}


void
alsa_patch_set_spec (alsa_patch_t * patch, const char * spec)
{
  snd_seq_addr_t from, to;
  char from_str[37];
  char to_str[37];
  int exclusive, time_update, time_real, queue, from_port, to_port;
  int err;
  
  err = sscanf (spec, "%s %d %s %d %d %d %d %d",
                from_str,
                &from_port,
                to_str,
                &to_port,
                &queue,
                &exclusive,
                &time_update,
                &time_real);
  
  if (err != 8)
    {
      fprintf (stderr, "%s: spec '%s' is not a valid alsa patch spec (err == %d)\n",
               __FUNCTION__, spec, err);
      return;
    }
  
  err = uuid_parse (from_str, patch->from_id);
  if (err == -1)
    {
      from.client = atoi (from_str);
       uuid_clear (patch->from_id);
    }
  from.port = from_port;
  snd_seq_port_subscribe_set_sender (patch->sub, &from);

  err = uuid_parse (to_str, patch->to_id);
  if (err == -1)
    {
      to.client = atoi (to_str);
      uuid_clear (patch->to_id);
    }
  to.port = to_port;
  snd_seq_port_subscribe_set_dest (patch->sub, &to);
  
  snd_seq_port_subscribe_set_queue (patch->sub, queue);
  snd_seq_port_subscribe_set_exclusive (patch->sub, exclusive);
  snd_seq_port_subscribe_set_time_update (patch->sub, time_update);
  snd_seq_port_subscribe_set_time_real (patch->sub, time_real);
  
}

void
alsa_patch_switch_clients (alsa_patch_t * patch)
{
  unsigned char from_client, from_port, to_client, to_port;
  snd_seq_addr_t addr;
  const snd_seq_addr_t * addrptr;
  
  addrptr = snd_seq_port_subscribe_get_sender (patch->sub);
  from_client = addrptr->client;
  from_port   = addrptr->port;

  addrptr = snd_seq_port_subscribe_get_dest (patch->sub);
  to_client = addrptr->client;
  to_port   = addrptr->port;
  
  addr.client = to_client;
  addr.port   = to_port;
  snd_seq_port_subscribe_set_sender (patch->sub, &addr);

  addr.client = from_client;
  addr.port   = from_port;
  snd_seq_port_subscribe_set_dest (patch->sub, &addr);
}


/* EOF */
